% ===================================================================
% Name: gBoost with label noise modelling
% Author: Jakramate Bootkrajang
% Date: Dec 2012
%
% Input: x = design matrix
%        y = class label
%        options.round = # of boosting rounds
%        options.g = label flipping matrix
%        options.estG  = estimating the label flipping matrix
%        options.verbose = plot some information
%        options.subsamp = adding diversity using subset resampling
%        options.learner = 'tree'(2 max splits) or 'rLR'
%        options.bType = Boosting type
%               : 1 is normal adaboost
%               : 2 is robust adaboost with g inside exp
%               : 3 is robust adaboost with g outside exp
%
% Return: a = weight vector of each member of the ensemble
%         h = a set of classifiers
% ===================================================================

function [a h] = gBoost(x, y, options,xt, yt)

% the function uses {-1,1} class label
y = castLabel(y,-1);

% check weak learner
if ~isfield(options, 'learner');
    disp('No weak learner specified, will use tree with 2 leaves');
    options.learner = 'tree';
end


% initialising hypothesis and its weight
switch options.learner
    case 'tree'
        weakLearner = true;      % tree is quite weak
        learner     = tree_node_w(2);
        % for tree
        h = {};
        a = [];
    case 'rlr'
        weakLearner  = false;    % rLR is quite strong
        % for rLR
        h = zeros(size(x,2)+1, options.round);
        a = zeros(1, options.round);
        % setting up options for rLR
        opts.boost   = true;
        opts.mUpdate = true;
        opts.verbose = false;
        opts.regFunc = 'noreg';
        options.subsamp = false;
    case 'lr'
        weakLearner  = false;    % LR is quite strong
        % for rLR
        h = zeros(size(x,2)+1, options.round);
        a = zeros(1, options.round);
        % setting up options for rLR
        opts.boost   = true;
        opts.mUpdate = true;
        opts.verbose = false;
        opts.regFunc = 'noreg';
        options.subsamp = true;
end

% initialising D to a uniform distribution
D = ones(size(x,1),1)./size(x,1);

% g table for gBoosting
g = options.gamma;

if strcmp(options.bType,'goada') == 1
    d11 = ones(sum(y== 1),1);
    d01 = ones(sum(y== 1),1);
    d00 = ones(sum(y==-1),1);
    d10 = ones(sum(y==-1),1);
    D(y== 1) = (g(1,1)*d11 - g(1,2)*d01);
    D(y==-1) = (g(2,2)*d00 - g(2,1)*d10);
end


% predefine some variables
idp  =  y == 1;
idn  =  y == -1;
opt.maxIter = 100;
opt.Display = false;

for t=1:options.round
    
    %% learning the base classifier
    if weakLearner
        % for tree classifier
        nodes = train(learner, x', y', D');
        final_hyp = 0;
        for i = 1:length(nodes)
            curr_tr = nodes{i};
            yHat = castLabel(calc_output(curr_tr, x'),-1);
            yHat = yHat';
            
            ep  =  sum(D .* (y== 1) .* (y ~= yHat));
            en  =  sum(D .* (y==-1) .* (y ~= yHat));
            et  =  ep + en;
            
            switch options.bType
                case 'ada'
                    ap =  0.5*log((1-et)/et);
                case 'giada'
                    [ap, fv, v1, v2] = minFunc(@grad_a_isd, 0, opt, D, y, g, ep, en);
                case 'goada'
                    [ap, fv, v1, v2] = minFunc(@grad_a_osd, 0, opt, d11, d01, d10, d00, g, et);
                otherwise
                    ap = 'ada';
            end
                        
            a(end+1) = ap;
            h{end+1} = curr_tr;
            
            final_hyp = final_hyp + yHat .* a(end);
        end
        % final hypothesis is the sum of all nodes of the current tree
        yHat = sign(final_hyp);
    else
        % for logistic regression or robust logistic regression
        if options.subsamp
            % subset resampling for creating diversity
            perm      = randperm(size(x,1));
            lim       = floor(size(x,1) * 0.3);
            X         = x(perm(1:lim),:);
            Y         = y(perm(1:lim));
            opts.dist = D(perm(1:lim));
            % setting g to identity results in traditional LR.
            gm        = [ 1  0;  0  1];
        else
            % using rLR to creating diversity
            X         = x;
            Y         = y;
            opts.dist = D;
            % setting g to something else results in robust LR.
            gm        = [.7 .3; .3 .7];
        end
        
        % learning the base classifier. *** rlr uses {0,1} class label ***
        % random initilisation of the weight vector is required.
        [h(:,t) gm llh] = rlr(randn(size(X,2)+1,1), gm, addbias(X), castLabel(Y,0), opts);
        yHat            = sign(addbias(x) * h(:,t));
    end
    
    %% evaluating the learner
    ep   =  sum(D .* (y== 1) .* (y ~= yHat));
    en   =  sum(D .* (y==-1) .* (y ~= yHat));
    et   =  ep + en;
    
    % if the error is too small stop
    if et < eps
        return;
    end    
    
    %% updating the distribution of the data
    switch options.bType
        case 'ada'
            %% discrete AdaBoost Ver.1  (The original)
            
            % finding optimal contribution of the current classifier
            alph =  0.5*log((1-et)/et);
            
            % update the distribution
            D     =  D .* exp(alph*(y ~= yHat));
            D     =  D / sum(D);
        case 'giada'
            %% discrete robust Adaboost with noise factors inside exp()
            % finding optimal contribution of the current classifier
            [alph, fv, v1, v2] = minFunc(@grad_a_isd, 0, opt, D, y, g, ep, en);
            
            % updating the distribution
            pen  =  zeros(size(y));
            idx  =  y ~= yHat;
            pen(idx & idn) = 1-(g(2,1));
            pen(idx & idp) = 1-(g(1,2));
            D  =  D .* exp(alph * pen);
            D  =  D / sum(D);
        case 'goada'
            %% discrete robust Adaboost with noise factors outside exp()
            % finding optimal contribution of the current classifier            
            [alph, fv, v1, v2] = minFunc(@grad_a_osd, 0, opt, d11, d01, d10, d00, g, et);
            %d = checkgrad('grad_a_osd', 5, 10-8, d11, d01, d10, d00, g, et)
            
            % updating the distribution, no normalisation applied here
            d11  =  d11 .* exp(alph * (yHat(idp) ~=  1)); 
            d01  =  d01 .* exp(alph * (yHat(idp) ==  1)); 
            d00  =  d00 .* exp(alph * (yHat(idn) ~= -1)); 
            d10  =  d10 .* exp(alph * (yHat(idn) == -1));                         
            
            D(idp) = (g(1,1)*d11 - g(1,2)*d01);
            D(idn) = (g(2,2)*d00 - g(2,1)*d10);
            
            % a negative weight is practically a zero weight            
            D(D<0) = 0;
        otherwise
            error('Please provide a type of boosting');
    end
    
    %% saving learner's contribution
    if weakLearner
        % for tree
        for i = 1:length(nodes)
            a(end+1-i) = a(end+1-i) * alph;
        end
    else
        % for rLR
        a(t) = alph;
    end
    
    %% updating the gamma matrix
    if options.estG && t > 10
        if weakLearner
            H = 0;
            for i = 1:size(a,2);
                pd = castLabel(calc_output(h{i}, x'),-1);
                H  = H + pd * a(i);
            end
            H = H';
        else
            pd  = sign(addbias(x) * h);
            H   = pd * a';
        end
        
        % 3 method of getting calibrated probability from boosting
        switch options.pType
            case 'logit';
                % logistic correction
                p1    = exp(H)./(exp(H)+exp(-H));
                p0    = 1 - p1;
            case 'platt';
                % platt's correction
                pd  = sign(addbias(xt) * h);
                Ht   = pd * a';
                [A B] = platt(Ht,yt,sum(yt==-1),sum(yt==1));
                p1    = 1./(1+exp(A*H+B));
                p0    = 1 - p1;
            case 'isotr';
                % isotonic regression
                p1    = monreggpav(H,castLabel(y,0),ones(size(y)));
                p0    = 1 - p1;
            otherwise
                error('Please provide probability calibration method');
        end
        
        s0  = g(1,1) * p0  +  g(2,1) * p1;
        s1  = g(1,2) * p0  +  g(2,2) * p1;
        
        % avoiding numerical problem
        s0(s0==0) = eps; s1(s1==0) = eps;
        
        g00 = g(1,1) * sum((y==-1) ./ s0 .* p0);
        g01 = g(1,2) * sum((y== 1) ./ s1 .* p0);        
        % (g00 + g01) is the Lagrangian
        g(1,1) = g00 / (g00 + g01);
        g(1,2) = 1 - g(1,1);
        
        g10 = g(2,1) * sum((y==-1) ./ s0 .* p1);
        g11 = g(2,2) * sum((y== 1) ./ s1 .* p1);        
        % (g10 + g11) is the Lagrangian
        g(2,1) = g10 / (g10 + g11);
        g(2,2) = 1 - g(2,1);

        %disp(g);
    end
    
    %% visualising the boosting process
    if options.verbose && t == options.round
        % (1) plotting the data set on subplot 1
        gridRange = -2.5:0.2:2.5;
        %subplot(1,2,1);
        plot(x(y==1,1),x(y==1,2),'kx',x(y==-1,1),x(y==-1,2),'ro','LineWidth',3,...
              'MarkerSize',6); hold on; drawnow
        set(gca,'FontSize',10);
        
%         % (2) plotting the decision boundary of the current learner on
%         % subplot 2
        [x1,x2] = meshgrid(gridRange, gridRange);
        xx      = [reshape(x1,size(x1,1)^2,1) reshape(x2,size(x2,1)^2,1)];
%         if weakLearner
%             [yHat v] = gBoostClassify(1, {h{end}}, xx, ones(size(xx,1),1));
%         else
%             yHat     = sign(addbias(xx) * h(:,t));
%         end
%         subplot(1,2,2);
%         if t>1; delete(hd); end;
%         [Cd hd] = contour(gridRange,gridRange,reshape(yHat,size(x1,1),size(x1,1)),1);
%         set(hd, 'LineWidth', 2, 'Color', [.0 .4 1]); drawnow
%         
        % (3) plotting the decision boundary of the ensemble on subplot 1
        if weakLearner
            [yHat v] = gBoostClassify(a(1:t), h, xx, ones(size(xx,1),1));
        else
            pd    = sign(addbias(xx) * h);
            H     = pd * a';
            yHat  = sign(H);  % final prediction
        end
        %subplot(1,2,1);
        [Cd hd] = contour(gridRange,gridRange,reshape(yHat,size(x1,1),size(x1,1)),1);
        set(hd, 'LineWidth', 5, 'Color', [.0 .4 1]); drawnow
    end    
    
    % norm of the estimated g to the true value
    % versus norm of the g to the identity matrix
    %plot(t, norm(eye(2)-g,'fro'),'bo',t,norm([.8 .2;0 1]-g,'fro'), 'rx'); hold on; drawnow
    
end  % end boosting loop

% % debugging
% if options.estG    
%     g
% end

